#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "shader.h"
#include "log.h"
#include "src/shader.h"

#define STATUS_INTERVAL 0.5

GLuint program;

GLuint vertexArrayObject;

GLuint initialWindowWidth = 800;
GLuint initialWindowHeight = 600;

int frameCount = 0;
struct timespec last_time, current_time;

void initialiseStatusDisplay() {
	clock_gettime(CLOCK_MONOTONIC, &last_time);
}

void updateStatusDisplay() {
	frameCount++;
	clock_gettime(CLOCK_MONOTONIC, &current_time);
	double elapsed = (current_time.tv_sec - last_time.tv_sec) +
					 (current_time.tv_nsec - last_time.tv_nsec) / 1e9;
	if (elapsed >= STATUS_INTERVAL) {
		double fps = frameCount / elapsed;
		frameCount = 0;
		last_time = current_time;
		printf("\rFPS: %.2f ", fps);
		fflush(stdout);
	}
}

void init(void) {
	INFO("Compiling Shaders...");

	// create and compile vertex shader
	INFO("Compiling Vertex Shader...");
	ShaderCompileResult vertexShader = readAndCompileShaderFromFile("src/shaders/vertex.glsl", GL_VERTEX_SHADER);

	if (!vertexShader.success) {
		FATAL("Failed to compile Vertex Shader");
		exit(1);
	}

	// create and compile fragment shader
	INFO("Compiling Fragment Shader...");
	ShaderCompileResult fragmentShader = readAndCompileShaderFromFile("src/shaders/fragment.glsl", GL_FRAGMENT_SHADER);
	if (!fragmentShader.success) {
		FATAL("Failed to compile Vertex Shader");
		exit(1);
	}

	// create and link shader program
	INFO("Linking Shader Program...");
	ProgramLinkResult linkResult = linkShaderProgram(vertexShader.shader, fragmentShader.shader);
	if (!linkResult.success) {
		FATAL("Failed to link Program");
		exit(1);
	}

	program = linkResult.program;

	INFO("Shader Program Done.");

	// create triangle buffer
	GLfloat triangleVertices[] =
	{
		 //X     //Y    // R  //G   /B
		 0.0f,  0.5f,   0.0f, 1.0f, 1.0f,
		-0.5f, -0.5f,   1.0f, 0.0f, 1.0f,
		 0.5f, -0.5f,   1.0f, 1.0f, 0.0f
	};

	DEBUG("Creating vertext buffer");
	GLuint vertexBuffer;
	glGenBuffers(1, &vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(triangleVertices), triangleVertices, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	DEBUG("Creating vertex array object");
	// create vertex array object
	glGenVertexArrays(1, &vertexArrayObject);
	glBindVertexArray(vertexArrayObject);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);

	// vertex position data
	glVertexAttribPointer(
		0, // shader location
		2, // number of values to read
		GL_FLOAT, // type of value
		GL_FALSE, // if values are normalised
		5 * sizeof(GLfloat), // stride - distance between vertices
		0 // start offset
	);
	glEnableVertexAttribArray(0);

	// vertex color data
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), (GLvoid*)(2 * sizeof(GLfloat)));
	glEnableVertexAttribArray(1);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);

	glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
	glViewport(0, 0, initialWindowWidth, initialWindowHeight);

	initialiseStatusDisplay();

	INFO("--- Initialisation done ---");
}

void draw(void) {
	updateStatusDisplay();
	glClear(GL_COLOR_BUFFER_BIT);
	glUseProgram(program);

	glBindVertexArray(vertexArrayObject);
	glDrawArrays(GL_TRIANGLES, 0, 3);
}

void framebuffer_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);
}

int main(int argc, char const *argv[])
{

	for (int i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-v") == 0) {
			logLevel = LOG_LEVEL_INFO;
		}
		if (strcmp(argv[i], "-vv") == 0) {
			logLevel = LOG_LEVEL_DEBUG;
		}
	}

	INFO("Creating GLFW Window");

	glfwInit();
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	GLFWwindow *window = glfwCreateWindow(initialWindowWidth, initialWindowHeight, "CG1", NULL, NULL);
	if (!window) {
		FATAL("Failed to open window");
		glfwTerminate();
		return 1;
	}
	glfwSetFramebufferSizeCallback(window, framebuffer_size_callback);
	glfwMakeContextCurrent(window);

	glewInit();

	init();

	while (!glfwWindowShouldClose(window)) {
		draw();

		glfwSwapBuffers(window);
		glfwPollEvents();
	}

	glfwTerminate();

	return 0;
}